/*****
 * xannotate.c
 * Annotate desktop with scribbles, for presentation.
 * Copyright (C) James Budiono, 2014
 * License: GNU GPL Version 3 or later.
 *****/
 
#include <X11/Xlib.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

//#define USE_SCREENSHOT // enable this to enable screenshot

#ifdef USE_SCREENSHOT
#include "pnglite.h"
#endif

#define DEFAULT_WIDTH   5
#define DEFAULT_HOTKEY "Pause"
#define DEFAULT_SCREENSHOT_PATH "xannotate"

// so we don't pass thousands of parameters each time
typedef struct
{
	Display *display;
	Window	root_window;
	int     screen;
	int     width, height;
	
	Window overlay;	
	Pixmap overlay_content, overlay_shapemask;
	GC content_red, content_black;
	GC bitmap_white, bitmap_black;
	unsigned int hotkey_keycode;
	
	int mouse_grabbed, visible;
	int do_draw, lastx, lasty;
	char *screenshot_path;
	int screenshot_index;
} ProgState;


// create the overlay window we will use as canvas
Window create_overlay_window (ProgState *st, Window *parent) 
{
	Window win;
	XSetWindowAttributes xswa;

	xswa.save_under = True;
	xswa.background_pixel = WhitePixel(st->display, st->screen);
	xswa.backing_store = NotUseful;
	xswa.override_redirect = True;
	xswa.cursor = XCreateFontCursor (st->display, XC_crosshair); 

	win = XCreateWindow(st->display, *parent, 0, 0, st->width, st->height,
		0, CopyFromParent, CopyFromParent, CopyFromParent,
		 CWBackPixel|CWBackingStore|CWSaveUnder|CWOverrideRedirect|CWCursor, &xswa);

	XSelectInput(st->display, win, StructureNotifyMask|ExposureMask);

	XMapWindow(st->display, win);

	return win;
}


// create GC (pen) for drawing, copying, etc
GC create_pen (ProgState *st, Drawable *d, char *colour, int width)
{
	GC pen;
	XGCValues values;
	Colormap cmap;
	XColor xc, xc2;

	cmap = DefaultColormap(st->display, st->screen);
	XAllocNamedColor(st->display, cmap, colour, &xc, &xc2);
	
	/* create the pen to draw lines with */
	values.line_width = width;
	values.line_style = LineSolid;
	values.foreground = xc.pixel;
	values.cap_style = CapRound;
	values.join_style = JoinMiter;
	return XCreateGC(st->display, *d, GCForeground|GCLineWidth|GCLineStyle|GCCapStyle|GCJoinStyle, &values);
}


// enable notification when our hotkey is pressed
unsigned int grab_hotkey (ProgState *st, char *hotkey) {
	unsigned int keycode;
	KeySym keysym;
	
	keysym = XStringToKeysym (hotkey);
	keycode = XKeysymToKeycode (st->display, keysym);
	if (keysym != NoSymbol && keycode) {
		XGrabKey (st->display, keycode, AnyModifier, st->root_window, False, GrabModeAsync, GrabModeAsync);
	}
	else 
	{
		fprintf (stderr, "I don't understand %s, bailing out.\n", hotkey);
		exit(1);
	}
	return keycode;
}


// grap or ungrab the mouse pointer
int grab_mouse (ProgState *st, int grab)
{
	if (grab) {
		if (!st->mouse_grabbed && st->visible) {
			fprintf (stderr, "mouse grab\n");
			XGrabPointer (st->display, st->overlay, False, 
			              ButtonPressMask|ButtonReleaseMask|PointerMotionMask,
			              GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
			st->mouse_grabbed = grab;
		}
	} else {
		if (st->mouse_grabbed) {
			fprintf (stderr, "mouse release\n");
			XUngrabPointer (st->display, CurrentTime);
			st->mouse_grabbed = grab;
		}		
	}
}


// show or hide window
void show_window (ProgState *st, int visible) {
	if (visible) {
		if (!st->visible) {
			fprintf (stderr, "show window\n");
			XMapWindow(st->display, st->overlay);
			st->visible = visible;
		}
	} else {
		if (st->visible) {
			fprintf (stderr, "hide window\n");
			grab_mouse (st, 0);
			XUnmapWindow(st->display, st->overlay);
			st->visible = visible;
		}		
	}
}


// update window content with latest drawing
void repaint (ProgState *st) {
	XCopyArea (st->display, st->overlay_content, st->overlay, st->content_black, 
	           0, 0, st->width, st->height, 0, 0);
	XShapeCombineMask(st->display, st->overlay, ShapeBounding, 0, 0, st->overlay_shapemask, ShapeSet);	
}


// draw a line
void draw_line (ProgState *st, int to_x, int to_y) {
	//fprintf (stderr, "draw %d, %d, %d, %d\n", st->lastx, st->lasty, to_x, to_y);
	XDrawLine(st->display, st->overlay_content, st->content_red, st->lastx, st->lasty, to_x, to_y);
	XDrawLine(st->display, st->overlay_shapemask, st->bitmap_white, st->lastx, st->lasty, to_x, to_y);
	repaint(st);
}


// clear canvas
void clear_drawing (ProgState *st) {
	fprintf (stderr, "clear drawing\n");
	XFillRectangle (st->display, st->overlay_content, st->content_black, 0, 0, st->width, st->height);
	XFillRectangle (st->display, st->overlay_shapemask, st->bitmap_black, 0, 0, st->width, st->height);	
}


// take screenshot (if enabled), using pnglite
void take_screenshot (ProgState *st) {
#ifdef USE_SCREENSHOT	
	XImage *image;
	png_t png;
	int x,y;
	unsigned char *buf;
	unsigned long pix;
	
	char *fname;
	struct stat sb;
	
	// find a file that does not exist yet
	fprintf (stderr, "screenshot\n");	y = 0;	
	do {
		asprintf (&fname, "%s%.5d.png", st->screenshot_path, st->screenshot_index++);
		if (stat (fname, &sb)) y=1;
	} while (!y);

	// do actual screenshot
	image = XGetImage (st->display, st->root_window, 0,0, st->width, st->height, AllPlanes, ZPixmap);
	buf = malloc (image->width * image->height * 3); // 3 bytes per pixel (=3 channels)
	for (y=0; y < image->height; y++)
	{
		for (x=0; x < image->width; x++)
		{
			pix = XGetPixel (image, x,y);
			buf[(y*image->width+x)*3+0] = (pix & image->red_mask)>>16;
			buf[(y*image->width+x)*3+1] = (pix & image->green_mask)>>8;
			buf[(y*image->width+x)*3+2] = (pix & image->blue_mask);
		}
	}
	png_open_file_write (&png, fname);
	png_set_data (&png, image->width, image->height, 8, PNG_TRUECOLOR, buf); //8 bits per channel.
	png_close_file (&png);
	free (buf);
	free (fname);
	XDestroyImage (image);
#else
	fprintf (stderr, "screenshot disabled\n");
#endif
}

// main event loop
int event_loop (ProgState *st)
{
	XEvent ev;
	
	/* as each event that we asked about occurs, we respond. */
	while(1){
		XNextEvent(st->display, &ev);
		
		switch(ev.type){
		case ButtonPress:
			st->do_draw = 1;
			st->lastx = ev.xbutton.x;
			st->lasty = ev.xbutton.y;
			//fprintf(stderr, "Press %d, %d, %d, %d, %d\n",ev.xbutton.window, ev.xbutton.x,ev.xbutton.y,ev.xbutton.x_root,ev.xbutton.y_root);
			break;

		case ButtonRelease:
			st->do_draw = 0;
			//fprintf(stderr, "Release %d, %d, %d, %d, %d\n",ev.xbutton.window, ev.xbutton.x,ev.xbutton.y,ev.xbutton.x_root,ev.xbutton.y_root);
			break;
		
		case MotionNotify:
			if (st->do_draw) {
				draw_line (st, ev.xmotion.x, ev.xmotion.y);
				st->lastx = ev.xmotion.x;
				st->lasty = ev.xmotion.y;
			}
			//fprintf(stderr, "Move %d, %d, %d, %d, %d\n",ev.xmotion.window, ev.xmotion.x,ev.xmotion.y,ev.xmotion.x_root,ev.xmotion.y_root);
			break;
			
		case Expose:
			repaint(st);
			break;

		case ConfigureNotify:
			//todo - if needed
			break;

		case KeymapNotify:
			XRefreshKeyboardMapping(&ev.xmapping);
			break;
			
		case KeyRelease:
			break; //ignore
			
		case KeyPress:
			{
				int state;		
				if (ev.xkey.keycode == st->hotkey_keycode) {
					state = 0;
					if (ev.xkey.state & ShiftMask) state ^= 1;
					if (ev.xkey.state & ControlMask) state ^= 2;
					if (ev.xkey.state & Mod1Mask) state ^= 4; // Alt
					switch (state) {
						case 0: // bare - mouse grab
							//fprintf(stderr, "Hotkey pressed, toggling mouse grab.\n");
							grab_mouse (st, !st->mouse_grabbed);
							break;
						case 1: // shift - show/hide
							//fprintf(stderr, "Shift-Hotkey pressed, toggle visibility.\n");
							show_window (st, !st->visible);
							break;
						case 2: // ctrl - clear
							//fprintf(stderr, "Ctrl-Hotkey pressed, clearing.\n");
							if (st->visible) {
								clear_drawing(st);
								repaint(st);
							}
							break;
						case 3: // ctrl-shift - exit 
							//fprintf(stderr, "Ctrl-Shift-Hotkey pressed, exiting.\n");
							XCloseDisplay(st->display);
							return 0;							
						case 4: // alt - screenshot
							//fprintf(stderr, "Alt-Hotkey pressed, screen capture.\n");
							take_screenshot(st);
							break;
					}
				}				
			}
			break;
		}
	}
}


//***************** Main ******************
int main (int argc, char*argv[])
{
	ProgState st;
	char *hotkey  = DEFAULT_HOTKEY;
	int pen_width = DEFAULT_WIDTH;
	int opt;

	// argument processing
	asprintf (&st.screenshot_path, "%s/xannotate", getenv("HOME"));
	st.screenshot_index = 0;
	while ((opt = getopt(argc, argv, "n:p:k:w:h")) != -1) 
	{
		switch (opt) 
		{
			case 'k':
				if (optarg) hotkey = optarg;
				break;
			case 'w':
				pen_width = atoi (optarg);
				if (!pen_width) pen_width = DEFAULT_WIDTH;
				break;
			case 'p':
				if (optarg) st.screenshot_path = optarg;
				break;
			case 'n':
				st.screenshot_index = atoi (optarg);
				break;
			case 'h':
				fprintf (stderr,
"Usage: xannotate [-k key] [-w width]\n"
"\t-k key:   use 'key' as the hotkey (default is 'Pause')\n"
"\t-w width: use 'width' as the default pen width (default is 5)\n"
"\t-p path:  path/filename for screenshot\n"
"\t-n index: the starting index used for screenshot filename\n");
				return 0;
		}
	}
	fprintf (stderr,
"--- Key usage ---\n"
"%s            - toggle draw/cursor mode\n"
"Shift-%s      - show/hide drawing\n"
"Ctrl-%s       - clear drawing\n"
"Ctrl-Shift-%s - exit\n"
"Alt-%s        - screenshot\n"
"---\n", hotkey, hotkey, hotkey, hotkey, hotkey);

	// initialisations
#ifdef USE_SCREENSHOT	
	png_init (0,0);
#endif
	st.mouse_grabbed = 0;
	st.do_draw = 0;
	st.visible = 1;	
	
    st.display = XOpenDisplay(NULL);
    if (st.display == NULL) {
        fprintf(stderr, "Cannot open display\n");
        exit(1);
    }
    st.screen            = DefaultScreen (st.display);
    st.root_window       = DefaultRootWindow (st.display);    
    st.width             = DisplayWidth(st.display, st.screen);
    st.height            = DisplayHeight(st.display, st.screen);
    st.hotkey_keycode    = grab_hotkey (&st, hotkey);

    st.overlay           = create_overlay_window (&st, &st.root_window);
    st.overlay_content   = XCreatePixmap (st.display, st.overlay, st.width, st.height, DefaultDepth(st.display, st.screen));
    st.overlay_shapemask = XCreatePixmap (st.display, st.overlay, st.width, st.height, 1);
    
    st.content_red       = create_pen (&st, &st.overlay_content, "Red", pen_width);
    st.content_black     = create_pen (&st, &st.overlay_content, "Black", pen_width);
    st.bitmap_black      = create_pen (&st, &st.overlay_shapemask, "Black", pen_width);
    st.bitmap_white      = create_pen (&st, &st.overlay_shapemask, "White", pen_width);
    
    clear_drawing (&st);
    return event_loop (&st);
}

